@@ -4,6 +4,11 @@ import os
 import threading
 from concurrent.futures import Future, ThreadPoolExecutor
 
+try:
+    import contextvars  # Python 3.7+ only.
+except ImportError:
+    contextvars = None
+
 
 class AsyncToSync:
     """
@@ -20,30 +25,43 @@ class AsyncToSync:
         except RuntimeError:
             # There's no event loop in this thread. Look for the threadlocal if
             # we're inside SyncToAsync
-            self.main_event_loop = getattr(SyncToAsync.threadlocal, "main_event_loop", None)
-            if self.main_event_loop is None:
-                raise RuntimeError(
-                    "You cannot instantiate AsyncToSync inside a thread that wasn't made using SyncToAsync"
-                )
+            self.main_event_loop = getattr(
+                SyncToAsync.threadlocal, "main_event_loop", None
+            )
 
     def __call__(self, *args, **kwargs):
+        # You can't call AsyncToSync from a thread with a running event loop
+        try:
+            event_loop = asyncio.get_event_loop()
+        except RuntimeError:
+            pass
+        else:
+            if event_loop.is_running():
+                raise RuntimeError(
+                    "You cannot use AsyncToSync in the same thread as an async event loop - "
+                    "just await the async function directly."
+                )
         # Make a future for the return information
         call_result = Future()
         # Use call_soon_threadsafe to schedule a synchronous callback on the
         # main event loop's thread
-        if not self.main_event_loop.is_running():
+        if not (self.main_event_loop and self.main_event_loop.is_running()):
             # Make our own event loop and run inside that.
             loop = asyncio.new_event_loop()
-            loop.run_until_complete(self.main_wrap(args, kwargs, call_result))
-            loop.close()
+            asyncio.set_event_loop(loop)
+            try:
+                loop.run_until_complete(self.main_wrap(args, kwargs, call_result))
+            finally:
+                try:
+                    if hasattr(loop, "shutdown_asyncgens"):
+                        loop.run_until_complete(loop.shutdown_asyncgens())
+                finally:
+                    loop.close()
+                    asyncio.set_event_loop(self.main_event_loop)
         else:
             self.main_event_loop.call_soon_threadsafe(
-                asyncio.ensure_future,
-                self.main_wrap(
-                    args,
-                    kwargs,
-                    call_result,
-                ),
+                self.main_event_loop.create_task,
+                self.main_wrap(args, kwargs, call_result),
             )
         # Wait for results from the future.
         return call_result.result()
@@ -74,13 +92,13 @@ class SyncToAsync:
     calls to AsyncToSync can escape it.
     """
 
-    threadpool = ThreadPoolExecutor(
-        max_workers=(
-            int(os.environ["ASGI_THREADS"])
-            if "ASGI_THREADS" in os.environ
-            else None
+    # If they've set ASGI_THREADS, update the default asyncio executor for now
+    if "ASGI_THREADS" in os.environ:
+        loop = asyncio.get_event_loop()
+        loop.set_default_executor(
+            ThreadPoolExecutor(max_workers=int(os.environ["ASGI_THREADS"]))
         )
-    )
+
     threadlocal = threading.local()
 
     def __init__(self, func):
@@ -88,9 +106,18 @@ class SyncToAsync:
 
     async def __call__(self, *args, **kwargs):
         loop = asyncio.get_event_loop()
+
+        if contextvars is not None:
+            context = contextvars.copy_context()
+            child = functools.partial(self.func, *args, **kwargs)
+            func = context.run
+            args = (child,)
+            kwargs = {}
+        else:
+            func = self.func
+
         future = loop.run_in_executor(
-            self.threadpool,
-            functools.partial(self.thread_handler, loop, *args, **kwargs),
+            None, functools.partial(self.thread_handler, loop, func, *args, **kwargs)
         )
         return await asyncio.wait_for(future, timeout=None)
 
@@ -100,14 +127,14 @@ class SyncToAsync:
         """
         return functools.partial(self.__call__, parent)
 
-    def thread_handler(self, loop, *args, **kwargs):
+    def thread_handler(self, loop, func, *args, **kwargs):
         """
         Wraps the sync application with exception handling.
         """
         # Set the threadlocal for AsyncToSync
         self.threadlocal.main_event_loop = loop
         # Run the function
-        return self.func(*args, **kwargs)
+        return func(*args, **kwargs)
 
 
 # Lowercase is more sensible for most things
